
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#pragma once

#include <glib-object.h>

G_BEGIN_DECLS

/* enumerations from "event.h" */

GType phoc_event_type_get_type (void);
#define PHOC_TYPE_EVENT_TYPE (phoc_event_type_get_type())

GType phoc_event_flags_get_type (void);
#define PHOC_TYPE_EVENT_FLAGS (phoc_event_flags_get_type())

GType phoc_touchpad_gesture_phase_get_type (void);
#define PHOC_TYPE_TOUCHPAD_GESTURE_PHASE (phoc_touchpad_gesture_phase_get_type())

GType phoc_event_sequence_state_get_type (void);
#define PHOC_TYPE_EVENT_SEQUENCE_STATE (phoc_event_sequence_state_get_type())

/* enumerations from "layer-shell-effects.h" */

GType phoc_draggable_surface_state_get_type (void);
#define PHOC_TYPE_DRAGGABLE_SURFACE_STATE (phoc_draggable_surface_state_get_type())

GType phoc_stacked_surface_pos_get_type (void);
#define PHOC_TYPE_STACKED_SURFACE_POS (phoc_stacked_surface_pos_get_type())

GType phoc_anim_dir_get_type (void);
#define PHOC_TYPE_ANIM_DIR (phoc_anim_dir_get_type())

/* enumerations from "output.h" */

GType phoc_output_scale_filter_get_type (void);
#define PHOC_TYPE_OUTPUT_SCALE_FILTER (phoc_output_scale_filter_get_type())

GType phoc_output_adaptive_sync_get_type (void);
#define PHOC_TYPE_OUTPUT_ADAPTIVE_SYNC (phoc_output_adaptive_sync_get_type())

/* enumerations from "phosh-private.h" */

GType phoc_phosh_private_shell_state_get_type (void);
#define PHOC_TYPE_PHOSH_PRIVATE_SHELL_STATE (phoc_phosh_private_shell_state_get_type())

/* enumerations from "server.h" */

GType phoc_server_flags_get_type (void);
#define PHOC_TYPE_SERVER_FLAGS (phoc_server_flags_get_type())

GType phoc_server_debug_flags_get_type (void);
#define PHOC_TYPE_SERVER_DEBUG_FLAGS (phoc_server_debug_flags_get_type())

/* enumerations from "view.h" */

GType phoc_view_tile_direction_get_type (void);
#define PHOC_TYPE_VIEW_TILE_DIRECTION (phoc_view_tile_direction_get_type())

GType phoc_view_state_get_type (void);
#define PHOC_TYPE_VIEW_STATE (phoc_view_state_get_type())

GType phoc_view_deco_part_get_type (void);
#define PHOC_TYPE_VIEW_DECO_PART (phoc_view_deco_part_get_type())

GType phoc_view_corner_get_type (void);
#define PHOC_TYPE_VIEW_CORNER (phoc_view_corner_get_type())

/* enumerations from "xwayland-unmanaged.h" */

GType phoc_xwayland_surface_state_get_type (void);
#define PHOC_TYPE_XWAYLAND_SURFACE_STATE (phoc_xwayland_surface_state_get_type())

G_END_DECLS

/* Generated data ends here */

