/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __PHOC_DBUS_DEBUG_CONTROL_H__
#define __PHOC_DBUS_DEBUG_CONTROL_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for mobi.phosh.Phoc.DebugControl */

#define PHOC_DBUS_TYPE_DEBUG_CONTROL (phoc_dbus_debug_control_get_type ())
#define PHOC_DBUS_DEBUG_CONTROL(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), PHOC_DBUS_TYPE_DEBUG_CONTROL, PhocDBusDebugControl))
#define PHOC_DBUS_IS_DEBUG_CONTROL(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), PHOC_DBUS_TYPE_DEBUG_CONTROL))
#define PHOC_DBUS_DEBUG_CONTROL_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), PHOC_DBUS_TYPE_DEBUG_CONTROL, PhocDBusDebugControlIface))

struct _PhocDBusDebugControl;
typedef struct _PhocDBusDebugControl PhocDBusDebugControl;
typedef struct _PhocDBusDebugControlIface PhocDBusDebugControlIface;

struct _PhocDBusDebugControlIface
{
  GTypeInterface parent_iface;

  gboolean  (*get_cutouts) (PhocDBusDebugControl *object);

  gboolean  (*get_damage_tracking) (PhocDBusDebugControl *object);

  gboolean  (*get_damage_whole) (PhocDBusDebugControl *object);

  const gchar *const * (*get_log_domains) (PhocDBusDebugControl *object);

  gboolean  (*get_touch_points) (PhocDBusDebugControl *object);

};

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (PhocDBusDebugControl, g_object_unref)
#endif

GType phoc_dbus_debug_control_get_type (void);

GDBusInterfaceInfo *phoc_dbus_debug_control_interface_info (void);
guint phoc_dbus_debug_control_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus property accessors: */
gboolean phoc_dbus_debug_control_get_cutouts (PhocDBusDebugControl *object);
void phoc_dbus_debug_control_set_cutouts (PhocDBusDebugControl *object, gboolean value);

gboolean phoc_dbus_debug_control_get_touch_points (PhocDBusDebugControl *object);
void phoc_dbus_debug_control_set_touch_points (PhocDBusDebugControl *object, gboolean value);

gboolean phoc_dbus_debug_control_get_damage_tracking (PhocDBusDebugControl *object);
void phoc_dbus_debug_control_set_damage_tracking (PhocDBusDebugControl *object, gboolean value);

gboolean phoc_dbus_debug_control_get_damage_whole (PhocDBusDebugControl *object);
void phoc_dbus_debug_control_set_damage_whole (PhocDBusDebugControl *object, gboolean value);

const gchar *const *phoc_dbus_debug_control_get_log_domains (PhocDBusDebugControl *object);
gchar **phoc_dbus_debug_control_dup_log_domains (PhocDBusDebugControl *object);
void phoc_dbus_debug_control_set_log_domains (PhocDBusDebugControl *object, const gchar *const *value);


/* ---- */

#define PHOC_DBUS_TYPE_DEBUG_CONTROL_PROXY (phoc_dbus_debug_control_proxy_get_type ())
#define PHOC_DBUS_DEBUG_CONTROL_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), PHOC_DBUS_TYPE_DEBUG_CONTROL_PROXY, PhocDBusDebugControlProxy))
#define PHOC_DBUS_DEBUG_CONTROL_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), PHOC_DBUS_TYPE_DEBUG_CONTROL_PROXY, PhocDBusDebugControlProxyClass))
#define PHOC_DBUS_DEBUG_CONTROL_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), PHOC_DBUS_TYPE_DEBUG_CONTROL_PROXY, PhocDBusDebugControlProxyClass))
#define PHOC_DBUS_IS_DEBUG_CONTROL_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), PHOC_DBUS_TYPE_DEBUG_CONTROL_PROXY))
#define PHOC_DBUS_IS_DEBUG_CONTROL_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), PHOC_DBUS_TYPE_DEBUG_CONTROL_PROXY))

typedef struct _PhocDBusDebugControlProxy PhocDBusDebugControlProxy;
typedef struct _PhocDBusDebugControlProxyClass PhocDBusDebugControlProxyClass;
typedef struct _PhocDBusDebugControlProxyPrivate PhocDBusDebugControlProxyPrivate;

struct _PhocDBusDebugControlProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  PhocDBusDebugControlProxyPrivate *priv;
};

struct _PhocDBusDebugControlProxyClass
{
  GDBusProxyClass parent_class;
};

GType phoc_dbus_debug_control_proxy_get_type (void);

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (PhocDBusDebugControlProxy, g_object_unref)
#endif

void phoc_dbus_debug_control_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
PhocDBusDebugControl *phoc_dbus_debug_control_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
PhocDBusDebugControl *phoc_dbus_debug_control_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void phoc_dbus_debug_control_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
PhocDBusDebugControl *phoc_dbus_debug_control_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
PhocDBusDebugControl *phoc_dbus_debug_control_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define PHOC_DBUS_TYPE_DEBUG_CONTROL_SKELETON (phoc_dbus_debug_control_skeleton_get_type ())
#define PHOC_DBUS_DEBUG_CONTROL_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), PHOC_DBUS_TYPE_DEBUG_CONTROL_SKELETON, PhocDBusDebugControlSkeleton))
#define PHOC_DBUS_DEBUG_CONTROL_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), PHOC_DBUS_TYPE_DEBUG_CONTROL_SKELETON, PhocDBusDebugControlSkeletonClass))
#define PHOC_DBUS_DEBUG_CONTROL_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), PHOC_DBUS_TYPE_DEBUG_CONTROL_SKELETON, PhocDBusDebugControlSkeletonClass))
#define PHOC_DBUS_IS_DEBUG_CONTROL_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), PHOC_DBUS_TYPE_DEBUG_CONTROL_SKELETON))
#define PHOC_DBUS_IS_DEBUG_CONTROL_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), PHOC_DBUS_TYPE_DEBUG_CONTROL_SKELETON))

typedef struct _PhocDBusDebugControlSkeleton PhocDBusDebugControlSkeleton;
typedef struct _PhocDBusDebugControlSkeletonClass PhocDBusDebugControlSkeletonClass;
typedef struct _PhocDBusDebugControlSkeletonPrivate PhocDBusDebugControlSkeletonPrivate;

struct _PhocDBusDebugControlSkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  PhocDBusDebugControlSkeletonPrivate *priv;
};

struct _PhocDBusDebugControlSkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType phoc_dbus_debug_control_skeleton_get_type (void);

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (PhocDBusDebugControlSkeleton, g_object_unref)
#endif

PhocDBusDebugControl *phoc_dbus_debug_control_skeleton_new (void);


G_END_DECLS

#endif /* __PHOC_DBUS_DEBUG_CONTROL_H__ */
