
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "phoc-enums.h"
#include "event.h"
#include "layer-shell-effects.h"
#include "output.h"
#include "phosh-private.h"
#include "server.h"
#include "view.h"
#include "xwayland-unmanaged.h"

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "event.h" */

GType
phoc_event_type_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (PHOC_EVENT_NOTHING), "PHOC_EVENT_NOTHING", "nothing" },
        { C_ENUM (PHOC_EVENT_BUTTON_PRESS), "PHOC_EVENT_BUTTON_PRESS", "button-press" },
        { C_ENUM (PHOC_EVENT_BUTTON_RELEASE), "PHOC_EVENT_BUTTON_RELEASE", "button-release" },
        { C_ENUM (PHOC_EVENT_MOTION_NOTIFY), "PHOC_EVENT_MOTION_NOTIFY", "motion-notify" },
        { C_ENUM (PHOC_EVENT_TOUCH_BEGIN), "PHOC_EVENT_TOUCH_BEGIN", "touch-begin" },
        { C_ENUM (PHOC_EVENT_TOUCH_UPDATE), "PHOC_EVENT_TOUCH_UPDATE", "touch-update" },
        { C_ENUM (PHOC_EVENT_TOUCH_END), "PHOC_EVENT_TOUCH_END", "touch-end" },
        { C_ENUM (PHOC_EVENT_TOUCH_CANCEL), "PHOC_EVENT_TOUCH_CANCEL", "touch-cancel" },
        { C_ENUM (PHOC_EVENT_TOUCHPAD_SWIPE_BEGIN), "PHOC_EVENT_TOUCHPAD_SWIPE_BEGIN", "touchpad-swipe-begin" },
        { C_ENUM (PHOC_EVENT_TOUCHPAD_SWIPE_UPDATE), "PHOC_EVENT_TOUCHPAD_SWIPE_UPDATE", "touchpad-swipe-update" },
        { C_ENUM (PHOC_EVENT_TOUCHPAD_SWIPE_END), "PHOC_EVENT_TOUCHPAD_SWIPE_END", "touchpad-swipe-end" },
        { C_ENUM (PHOC_EVENT_TOUCHPAD_PINCH_BEGIN), "PHOC_EVENT_TOUCHPAD_PINCH_BEGIN", "touchpad-pinch-begin" },
        { C_ENUM (PHOC_EVENT_TOUCHPAD_PINCH_UPDATE), "PHOC_EVENT_TOUCHPAD_PINCH_UPDATE", "touchpad-pinch-update" },
        { C_ENUM (PHOC_EVENT_TOUCHPAD_PINCH_END), "PHOC_EVENT_TOUCHPAD_PINCH_END", "touchpad-pinch-end" },
        { C_ENUM (PHOC_EVENT_GRAB_BROKEN), "PHOC_EVENT_GRAB_BROKEN", "grab-broken" },
        { C_ENUM (PHOC_EVENT_EVENT_LAST), "PHOC_EVENT_EVENT_LAST", "event-last" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("PhocEventType"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
phoc_event_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (PHOC_EVENT_NONE), "PHOC_EVENT_NONE", "none" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("PhocEventFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
phoc_touchpad_gesture_phase_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (PHOC_TOUCHPAD_GESTURE_PHASE_BEGIN), "PHOC_TOUCHPAD_GESTURE_PHASE_BEGIN", "begin" },
        { C_ENUM (PHOC_TOUCHPAD_GESTURE_PHASE_UPDATE), "PHOC_TOUCHPAD_GESTURE_PHASE_UPDATE", "update" },
        { C_ENUM (PHOC_TOUCHPAD_GESTURE_PHASE_END), "PHOC_TOUCHPAD_GESTURE_PHASE_END", "end" },
        { C_ENUM (PHOC_TOUCHPAD_GESTURE_PHASE_CANCEL), "PHOC_TOUCHPAD_GESTURE_PHASE_CANCEL", "cancel" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("PhocTouchpadGesturePhase"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
phoc_event_sequence_state_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (PHOC_EVENT_SEQUENCE_NONE), "PHOC_EVENT_SEQUENCE_NONE", "none" },
        { C_ENUM (PHOC_EVENT_SEQUENCE_CLAIMED), "PHOC_EVENT_SEQUENCE_CLAIMED", "claimed" },
        { C_ENUM (PHOC_EVENT_SEQUENCE_DENIED), "PHOC_EVENT_SEQUENCE_DENIED", "denied" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("PhocEventSequenceState"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "layer-shell-effects.h" */

GType
phoc_draggable_surface_state_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (PHOC_DRAGGABLE_SURFACE_STATE_NONE), "PHOC_DRAGGABLE_SURFACE_STATE_NONE", "none" },
        { C_ENUM (PHOC_DRAGGABLE_SURFACE_STATE_PENDING), "PHOC_DRAGGABLE_SURFACE_STATE_PENDING", "pending" },
        { C_ENUM (PHOC_DRAGGABLE_SURFACE_STATE_DRAGGING), "PHOC_DRAGGABLE_SURFACE_STATE_DRAGGING", "dragging" },
        { C_ENUM (PHOC_DRAGGABLE_SURFACE_STATE_ANIMATING), "PHOC_DRAGGABLE_SURFACE_STATE_ANIMATING", "animating" },
        { C_ENUM (PHOC_DRAGGABLE_SURFACE_STATE_REJECTED), "PHOC_DRAGGABLE_SURFACE_STATE_REJECTED", "rejected" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("PhocDraggableSurfaceState"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
phoc_stacked_surface_pos_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (PHOC_STACKED_SURFACE_STACK_BELOW), "PHOC_STACKED_SURFACE_STACK_BELOW", "below" },
        { C_ENUM (PHOC_STACKED_SURFACE_STACK_ABOVE), "PHOC_STACKED_SURFACE_STACK_ABOVE", "above" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("PhocStackedSurfacePos"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
phoc_anim_dir_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (ANIM_DIR_IN), "ANIM_DIR_IN", "in" },
        { C_ENUM (ANIM_DIR_OUT), "ANIM_DIR_OUT", "out" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("PhocAnimDir"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "output.h" */

GType
phoc_output_scale_filter_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (PHOC_OUTPUT_SCALE_FILTER_AUTO), "PHOC_OUTPUT_SCALE_FILTER_AUTO", "auto" },
        { C_ENUM (PHOC_OUTPUT_SCALE_FILTER_BILINEAR), "PHOC_OUTPUT_SCALE_FILTER_BILINEAR", "bilinear" },
        { C_ENUM (PHOC_OUTPUT_SCALE_FILTER_NEAREST), "PHOC_OUTPUT_SCALE_FILTER_NEAREST", "nearest" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("PhocOutputScaleFilter"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
phoc_output_adaptive_sync_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (PHOC_OUTPUT_ADAPTIVE_SYNC_NONE), "PHOC_OUTPUT_ADAPTIVE_SYNC_NONE", "none" },
        { C_ENUM (PHOC_OUTPUT_ADAPTIVE_SYNC_ENABLED), "PHOC_OUTPUT_ADAPTIVE_SYNC_ENABLED", "enabled" },
        { C_ENUM (PHOC_OUTPUT_ADAPTIVE_SYNC_DISABLED), "PHOC_OUTPUT_ADAPTIVE_SYNC_DISABLED", "disabled" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("PhocOutputAdaptiveSync"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "phosh-private.h" */

GType
phoc_phosh_private_shell_state_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (PHOC_PHOSH_PRIVATE_SHELL_STATE_UNKNOWN), "PHOC_PHOSH_PRIVATE_SHELL_STATE_UNKNOWN", "unknown" },
        { C_ENUM (PHOC_PHOSH_PRIVATE_SHELL_STATE_UP), "PHOC_PHOSH_PRIVATE_SHELL_STATE_UP", "up" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("PhocPhoshPrivateShellState"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "server.h" */

GType
phoc_server_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (PHOC_SERVER_FLAG_NONE), "PHOC_SERVER_FLAG_NONE", "none" },
        { C_FLAGS (PHOC_SERVER_FLAG_SHELL_MODE), "PHOC_SERVER_FLAG_SHELL_MODE", "shell-mode" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("PhocServerFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
phoc_server_debug_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (PHOC_SERVER_DEBUG_FLAG_NONE), "PHOC_SERVER_DEBUG_FLAG_NONE", "none" },
        { C_FLAGS (PHOC_SERVER_DEBUG_FLAG_AUTO_MAXIMIZE), "PHOC_SERVER_DEBUG_FLAG_AUTO_MAXIMIZE", "auto-maximize" },
        { C_FLAGS (PHOC_SERVER_DEBUG_FLAG_DAMAGE_TRACKING), "PHOC_SERVER_DEBUG_FLAG_DAMAGE_TRACKING", "damage-tracking" },
        { C_FLAGS (PHOC_SERVER_DEBUG_FLAG_NO_QUIT), "PHOC_SERVER_DEBUG_FLAG_NO_QUIT", "no-quit" },
        { C_FLAGS (PHOC_SERVER_DEBUG_FLAG_TOUCH_POINTS), "PHOC_SERVER_DEBUG_FLAG_TOUCH_POINTS", "touch-points" },
        { C_FLAGS (PHOC_SERVER_DEBUG_FLAG_LAYER_SHELL), "PHOC_SERVER_DEBUG_FLAG_LAYER_SHELL", "layer-shell" },
        { C_FLAGS (PHOC_SERVER_DEBUG_FLAG_CUTOUTS), "PHOC_SERVER_DEBUG_FLAG_CUTOUTS", "cutouts" },
        { C_FLAGS (PHOC_SERVER_DEBUG_FLAG_DISABLE_ANIMATIONS), "PHOC_SERVER_DEBUG_FLAG_DISABLE_ANIMATIONS", "disable-animations" },
        { C_FLAGS (PHOC_SERVER_DEBUG_FLAG_FORCE_SHELL_REVEAL), "PHOC_SERVER_DEBUG_FLAG_FORCE_SHELL_REVEAL", "force-shell-reveal" },
        { C_FLAGS (PHOC_SERVER_DEBUG_FLAG_IGNORE_STATES), "PHOC_SERVER_DEBUG_FLAG_IGNORE_STATES", "ignore-states" },
        { C_FLAGS (PHOC_SERVER_DEBUG_FLAG_DAMAGE_WHOLE), "PHOC_SERVER_DEBUG_FLAG_DAMAGE_WHOLE", "damage-whole" },
        { C_FLAGS (PHOC_SERVER_DEBUG_FLAG_FAKE_BUILTIN), "PHOC_SERVER_DEBUG_FLAG_FAKE_BUILTIN", "fake-builtin" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("PhocServerDebugFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "view.h" */

GType
phoc_view_tile_direction_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (PHOC_VIEW_TILE_NONE), "PHOC_VIEW_TILE_NONE", "none" },
        { C_FLAGS (PHOC_VIEW_TILE_LEFT), "PHOC_VIEW_TILE_LEFT", "left" },
        { C_FLAGS (PHOC_VIEW_TILE_RIGHT), "PHOC_VIEW_TILE_RIGHT", "right" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("PhocViewTileDirection"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
phoc_view_state_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (PHOC_VIEW_STATE_FLOATING), "PHOC_VIEW_STATE_FLOATING", "floating" },
        { C_ENUM (PHOC_VIEW_STATE_MAXIMIZED), "PHOC_VIEW_STATE_MAXIMIZED", "maximized" },
        { C_ENUM (PHOC_VIEW_STATE_TILED), "PHOC_VIEW_STATE_TILED", "tiled" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("PhocViewState"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
phoc_view_deco_part_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (PHOC_VIEW_DECO_PART_NONE), "PHOC_VIEW_DECO_PART_NONE", "none" },
        { C_FLAGS (PHOC_VIEW_DECO_PART_TOP_BORDER), "PHOC_VIEW_DECO_PART_TOP_BORDER", "top-border" },
        { C_FLAGS (PHOC_VIEW_DECO_PART_BOTTOM_BORDER), "PHOC_VIEW_DECO_PART_BOTTOM_BORDER", "bottom-border" },
        { C_FLAGS (PHOC_VIEW_DECO_PART_LEFT_BORDER), "PHOC_VIEW_DECO_PART_LEFT_BORDER", "left-border" },
        { C_FLAGS (PHOC_VIEW_DECO_PART_RIGHT_BORDER), "PHOC_VIEW_DECO_PART_RIGHT_BORDER", "right-border" },
        { C_FLAGS (PHOC_VIEW_DECO_PART_TITLEBAR), "PHOC_VIEW_DECO_PART_TITLEBAR", "titlebar" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("PhocViewDecoPart"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
phoc_view_corner_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (PHOC_VIEW_CORNER_NORTH_WEST), "PHOC_VIEW_CORNER_NORTH_WEST", "north-west" },
        { C_ENUM (PHOC_VIEW_CORNER_NORTH_EAST), "PHOC_VIEW_CORNER_NORTH_EAST", "north-east" },
        { C_ENUM (PHOC_VIEW_CORNER_SOUTH_EAST), "PHOC_VIEW_CORNER_SOUTH_EAST", "south-east" },
        { C_ENUM (PHOC_VIEW_CORNER_SOUTH_WEST), "PHOC_VIEW_CORNER_SOUTH_WEST", "south-west" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("PhocViewCorner"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "xwayland-unmanaged.h" */

GType
phoc_xwayland_surface_state_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (PHOC_XWAYLAND_SURFACE_STATE_NONE), "PHOC_XWAYLAND_SURFACE_STATE_NONE", "none" },
        { C_FLAGS (PHOC_XWAYLAND_SURFACE_STATE_MAPPED), "PHOC_XWAYLAND_SURFACE_STATE_MAPPED", "mapped" },
        { C_FLAGS (PHOC_XWAYLAND_SURFACE_STATE_ASSOCIATED), "PHOC_XWAYLAND_SURFACE_STATE_ASSOCIATED", "associated" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("PhocXWaylandSurfaceState"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* Generated data ends here */

