
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "phoc-anim-enum-types.h"
#include "phoc-anim-enums.h"

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "phoc-anim-enums.h" */

GType
phoc_animation_state_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (PHOC_TIMED_ANIMATION_IDLE), "PHOC_TIMED_ANIMATION_IDLE", "idle" },
        { C_ENUM (PHOC_TIMED_ANIMATION_PLAYING), "PHOC_TIMED_ANIMATION_PLAYING", "playing" },
        { C_ENUM (PHOC_TIMED_ANIMATION_FINISHED), "PHOC_TIMED_ANIMATION_FINISHED", "finished" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("PhocAnimationState"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
phoc_easing_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (PHOC_EASING_NONE), "PHOC_EASING_NONE", "none" },
        { C_ENUM (PHOC_EASING_EASE_IN_QUAD), "PHOC_EASING_EASE_IN_QUAD", "ease-in-quad" },
        { C_ENUM (PHOC_EASING_EASE_OUT_QUAD), "PHOC_EASING_EASE_OUT_QUAD", "ease-out-quad" },
        { C_ENUM (PHOC_EASING_EASE_IN_OUT_QUAD), "PHOC_EASING_EASE_IN_OUT_QUAD", "ease-in-out-quad" },
        { C_ENUM (PHOC_EASING_EASE_IN_CUBIC), "PHOC_EASING_EASE_IN_CUBIC", "ease-in-cubic" },
        { C_ENUM (PHOC_EASING_EASE_OUT_CUBIC), "PHOC_EASING_EASE_OUT_CUBIC", "ease-out-cubic" },
        { C_ENUM (PHOC_EASING_EASE_IN_OUT_CUBIC), "PHOC_EASING_EASE_IN_OUT_CUBIC", "ease-in-out-cubic" },
        { C_ENUM (PHOC_EASING_EASE_IN_QUART), "PHOC_EASING_EASE_IN_QUART", "ease-in-quart" },
        { C_ENUM (PHOC_EASING_EASE_OUT_QUART), "PHOC_EASING_EASE_OUT_QUART", "ease-out-quart" },
        { C_ENUM (PHOC_EASING_EASE_IN_OUT_QUART), "PHOC_EASING_EASE_IN_OUT_QUART", "ease-in-out-quart" },
        { C_ENUM (PHOC_EASING_EASE_IN_QUINT), "PHOC_EASING_EASE_IN_QUINT", "ease-in-quint" },
        { C_ENUM (PHOC_EASING_EASE_OUT_QUINT), "PHOC_EASING_EASE_OUT_QUINT", "ease-out-quint" },
        { C_ENUM (PHOC_EASING_EASE_IN_OUT_QUINT), "PHOC_EASING_EASE_IN_OUT_QUINT", "ease-in-out-quint" },
        { C_ENUM (PHOC_EASING_EASE_IN_SINE), "PHOC_EASING_EASE_IN_SINE", "ease-in-sine" },
        { C_ENUM (PHOC_EASING_EASE_OUT_SINE), "PHOC_EASING_EASE_OUT_SINE", "ease-out-sine" },
        { C_ENUM (PHOC_EASING_EASE_IN_OUT_SINE), "PHOC_EASING_EASE_IN_OUT_SINE", "ease-in-out-sine" },
        { C_ENUM (PHOC_EASING_EASE_IN_EXPO), "PHOC_EASING_EASE_IN_EXPO", "ease-in-expo" },
        { C_ENUM (PHOC_EASING_EASE_OUT_EXPO), "PHOC_EASING_EASE_OUT_EXPO", "ease-out-expo" },
        { C_ENUM (PHOC_EASING_EASE_IN_OUT_EXPO), "PHOC_EASING_EASE_IN_OUT_EXPO", "ease-in-out-expo" },
        { C_ENUM (PHOC_EASING_EASE_IN_CIRC), "PHOC_EASING_EASE_IN_CIRC", "ease-in-circ" },
        { C_ENUM (PHOC_EASING_EASE_OUT_CIRC), "PHOC_EASING_EASE_OUT_CIRC", "ease-out-circ" },
        { C_ENUM (PHOC_EASING_EASE_IN_OUT_CIRC), "PHOC_EASING_EASE_IN_OUT_CIRC", "ease-in-out-circ" },
        { C_ENUM (PHOC_EASING_EASE_IN_ELASTIC), "PHOC_EASING_EASE_IN_ELASTIC", "ease-in-elastic" },
        { C_ENUM (PHOC_EASING_EASE_OUT_ELASTIC), "PHOC_EASING_EASE_OUT_ELASTIC", "ease-out-elastic" },
        { C_ENUM (PHOC_EASING_EASE_IN_OUT_ELASTIC), "PHOC_EASING_EASE_IN_OUT_ELASTIC", "ease-in-out-elastic" },
        { C_ENUM (PHOC_EASING_EASE_IN_BACK), "PHOC_EASING_EASE_IN_BACK", "ease-in-back" },
        { C_ENUM (PHOC_EASING_EASE_OUT_BACK), "PHOC_EASING_EASE_OUT_BACK", "ease-out-back" },
        { C_ENUM (PHOC_EASING_EASE_IN_OUT_BACK), "PHOC_EASING_EASE_IN_OUT_BACK", "ease-in-out-back" },
        { C_ENUM (PHOC_EASING_EASE_IN_BOUNCE), "PHOC_EASING_EASE_IN_BOUNCE", "ease-in-bounce" },
        { C_ENUM (PHOC_EASING_EASE_OUT_BOUNCE), "PHOC_EASING_EASE_OUT_BOUNCE", "ease-out-bounce" },
        { C_ENUM (PHOC_EASING_EASE_IN_OUT_BOUNCE), "PHOC_EASING_EASE_IN_OUT_BOUNCE", "ease-in-out-bounce" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("PhocEasing"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* Generated data ends here */

