/*
 * Copyright (C) 2024 The Phosh Developers
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#include "desktop.h"
#include "xwayland-surface.h"

G_BEGIN_DECLS

void phoc_desktop_setup_xwayland   (PhocDesktop *self);
void phoc_desktop_destroy_xwayland (PhocDesktop *self);

G_END_DECLS
